# -*- coding: utf-8 -*-
"""
Created on Mon Jul 24 11:27:33 2023

@author: Ryan
"""


from pykrige.ok import OrdinaryKriging
from pykrige.uk import UniversalKriging
import numpy as np   
import copy  
from harpia.tropomi.no2 import kriging

def make_mask(data_array, predict_indices):
    
    mask=np.ones(data_array.shape)
    nan_indices=np.where(np.isnan(data_array))
    
    mask[nan_indices]=-1
    
    #negative_indices=np.where(data_array<0)
    #mask[negative_indices]=-1
    
    mask[predict_indices]=0
    
    
    return mask
##############################################################################
##############################################################################
##############################################################################

def kriging_helper(array,mask,kriging_type, variogram_model):
    
    
    one_indices=np.where(mask==1)
    y_array=one_indices[0]
    x_array=one_indices[1]
    
    
    
    k_y=y_array.tolist()
    k_x=x_array.tolist()
    
    data=array[one_indices]
    k_data=data.tolist()
    if kriging_type=='ordinary':
        k_model=OrdinaryKriging(k_x,k_y,k_data,variogram_model)
    elif kriging_type=='universal':
        k_model=UniversalKriging(k_x,k_y,k_data,variogram_model)
    else:
        raise "only universal or ordinary kriging are supported"
    zero_indices=np.where(mask==0)

    y_predict_array=zero_indices[0]
    x_predict_array=zero_indices[1]
    
    yy=y_predict_array.astype('float64')
    xx=x_predict_array.astype('float64')
    k_p_y=yy.tolist()
    k_p_x=xx.tolist()
    
    z,ss=k_model.execute('points',k_p_x,k_p_y)
   
    return z

##############################################################################
##############################################################################
##############################################################################
  
def main_kriging(data, predict_indices,kriging_type, variogram_model): #data must have negative and nan values as they are
    
    
    mask=kriging.make_mask(data,predict_indices)
    
    #data_array=(data-np.nanmin(data))/(np.nanmax(data)-np.nanmin(data))
    k_data=copy.deepcopy(data)
    
    for i in range(data.shape[0]):
        slice_mask=mask[i,:,:]
        slice_data=k_data[i,:,:]
        indices=np.where(slice_mask==0)
        
        z=kriging.kriging_helper(slice_data,slice_mask,kriging_type, variogram_model)
        slice_data[indices]=z
        k_data[i,:,:]=slice_data
        print(i)
    
    #k_data=k_data*(np.nanmax(data)-np.nanmin(data))+np.nanmin(data)
    return k_data[predict_indices]
##############################################################################
##############################################################################
##############################################################################
def run_kriging_helper(data,random_mask,row_num, column_num,row_size, column_size,kriging_type, variogram_model):
    k_data=copy.deepcopy(data)
    for i in range(0,row_num):
        for j in range(0,column_num):
            data_array=data[:,(row_size*i):(row_size*(i+1)),(column_size*j):(column_size*(j+1))]
            mask_array=random_mask[:,(row_size*i):(row_size*(i+1)),(column_size*j):(column_size*(j+1))]
            
            random_indices=np.where(mask_array==0)
            z=kriging.main_kriging(data_array.copy(),random_indices,kriging_type,variogram_model)
            
            data_array[random_indices]=z
            
            
            k_data[:,(row_size*i):(row_size*(i+1)),(column_size*j):(column_size*(j+1))]=data_array
        #print(i)
    return k_data  

    
def run_kriging_helper_custom(data,random_mask,rows, cols, kriging_type, variogram_model):
    k_data=copy.deepcopy(data)
    for i in range(len(rows)-1):
        for j in range(len(cols)-1):
            data_array=data[:,rows[i]:rows[i+1],cols[j]:cols[j+1]]
            mask_array=random_mask[:,rows[i]:rows[i+1],cols[j]:cols[j+1]]
    
            random_indices=np.where(mask_array==0)
            z=kriging.main_kriging(data_array.copy(),random_indices,kriging_type,variogram_model)
            
            data_array[random_indices]=z
            k_data[:,rows[i]:rows[i+1],cols[j]:cols[j+1]]=data_array
            
    return k_data
##############################################################################
##############################################################################
##############################################################################


def run_kriging(data,random_mask,resolution, kriging_type, variogram_model, partitioning=False):
    if resolution=='050':
        k_data=kriging.run_kriging_helper(data,random_mask,1,1,70,140,kriging_type, variogram_model)
    elif resolution=='025':
        if partitioning:
            k_data=kriging.run_kriging_helper(data,random_mask,3,1,47,281,kriging_type,variogram_model)
        else:
            k_data=kriging.run_kriging_helper(data,random_mask,1,1,141,281,kriging_type,variogram_model)
    else:
        raise "resolution is not recognized"
    return k_data


##############################################################################

def run_custom_kriging(data,random_mask,resolution, rows, cols, kriging_type, variogram_model): #rows and cols determines the partitioning
    k_data=kriging.run_kriging_helper_custom(data,random_mask,rows, cols, kriging_type, variogram_model)
    return k_data
    

    
    



# -*- coding: utf-8 -*-
"""
Created on Thu Jul 20 12:01:52 2023

@author: Ryan
"""



from datetime import datetime
from datetime import timedelta
import rasterio
import os
import numpy as np
import pkg_resources
from harpia.tropomi.no2 import utils

##############################################################################
##############################################################################
def generate_date_list(start_date,end_date):  
    
    date=datetime.strptime(start_date,'%Y%m%d')
    stop_date=datetime.strptime(end_date,'%Y%m%d')
    granule_end_date_list=[]

    while date<stop_date:
        granule_end_date=date.date().strftime('%Y%m%d')
        granule_end_date_list.append(granule_end_date)
        
        date+=timedelta(days=1)
    return granule_end_date_list
##############################################################################
##############################################################################
##############################################################################

def raster_to_array(raster_directory, start_date, end_date, resolution, qa_value):

    date_list=utils.generate_date_list(start_date,end_date)

    data=[]
    existed_list=[]
    missed_dates=[]
    for date in date_list:
        
        file_name=date+'_'+resolution+'_'+qa_value+'_'+'raster.tif'
        
        file_path=os.path.join(raster_directory,file_name)
        
        if os.path.exists(file_path):
        
            file_data=rasterio.open(file_path)
        
            array=file_data.read(1)
        
            data.append(array)
            
            existed_list.append(date)
        else:
            missed_dates.append(date)
        
    data_array=np.array(data)
    

    return data_array, existed_list, missed_dates
##############################################################################
##############################################################################
##############################################################################

def add_nans(data):
    
    data_array=data.copy()
    
    nan_indices=np.where(data_array==-1E+300) #the fill value for rasterization is replaced by nan
    data_array[nan_indices]=np.nan
    
    return data_array
##############################################################################
##############################################################################
##############################################################################
def molec_per_cm2(data):
    
    data_array=data.copy()
    real_indices=np.where(np.logical_not(np.isnan(data_array)))
    data_array[real_indices]=data_array[real_indices]*6.02214E+19
    return data_array
##############################################################################
##############################################################################
##############################################################################


def make_mask(data_array):
    
    mask=np.ones(data_array.shape)
    nan_indices=np.where(np.isnan(data_array))
    
    mask[nan_indices]=0
    
    #negative_indices=np.where(data_array<0)
    #mask[negative_indices]=0
    
    
    return mask
##############################################################################
##############################################################################
############################################################################## 

def generate_random_indices(mask,n):
    
    one_indices=np.where(mask==1)
    random_pos=np.random.choice(len(one_indices[0]),n, replace=False)
    array_one_indices=np.vstack(one_indices)
    
    
    random_init_list=[]
    for i in random_pos:
        random_init_list.append(array_one_indices[:,i])
    
    random_indices_array=np.vstack(random_init_list).T
    
    
    random_indices=[]
    for r in range(random_indices_array.shape[0]):
        random_indices.append(random_indices_array[r])
    
    
    return tuple(random_indices)
##############################################################################

##############################################################################
def read_added_missing_indices(file_name):
    file_path=os.path.join('lib/random_missing_indices', file_name)
    with pkg_resources.resource_stream(__name__, file_path) as stream:
        
        random_indices_array=np.load(stream)
        random_indices=tuple(random_indices_array.copy())
        
    return random_indices



   
##############################################################################
##############################################################################
##############################################################################

def index_of_agreement(observations,predictions): #Willmott 1981

    num=np.sum(np.square(observations-predictions))
    o_mean=np.mean(observations)
    denom=np.sum((abs(predictions-o_mean)+abs(observations-o_mean))**2)
    
    ioa=1-(num/denom)
    return ioa

def modified_index_of_agreement(observations,predictions): #Mccabe Jr (1999)
    num=np.sum(abs(observations-predictions))
    o_mean=np.mean(observations)
    denom=np.sum((abs(predictions-o_mean)+abs(observations-o_mean)))
    
    ioa=1-(num/denom)
    return ioa   

def mean_absolute_error(observations,predictions):
    return np.mean(abs(observations-predictions))


def root_mean_square_error(observations,predictions):
    return np.mean(np.square(observations-predictions))**(0.5)

def correlation_coefficient(observations,predictions):
    return np.corrcoef(observations,predictions)[0][1]

def mean_absolute_percentage_error(observations,predictions):
    return np.mean(abs((observations-predictions)/observations))



    


 # -*- coding: utf-8 -*-
"""
Created on Sun Jul  9 20:20:41 2023

@author: Ryan
"""


import os
import numpy as np
from netCDF4 import Dataset
import pandas as pd
import geopandas
from rasterio.features import rasterize
import rasterio
import pkg_resources
from harpia.tropomi.no2 import converter

def sat_to_df(data_file_path):
    
    file=Dataset(data_file_path, "r" , format="NETCDF4")
    
    
    qa_value=file.groups['PRODUCT'].variables['qa_value']
    tropospheric_no2=file.groups['PRODUCT'].variables['nitrogendioxide_tropospheric_column']
    latitude=file.groups['PRODUCT'].variables['latitude']
    longitude=file.groups['PRODUCT'].variables['longitude']

    

    qa_value_array=np.array(qa_value)
    tropospheric_no2_array=np.array(tropospheric_no2)
    latitude_array=np.array(latitude)
    longitude_array=np.array(longitude)
    
    #tropospheric_no2_array=(tropospheric_no2_array-np.min(tropospheric_no2_array))/(np.max(tropospheric_no2_array)-np.min(tropospheric_no2_array))
    
    
    lat_flat=latitude_array.flatten()
    long_flat=longitude_array.flatten()
    no2_flat=tropospheric_no2_array.flatten()
    qa_value_flat=qa_value_array.flatten()
    
    
    df_dict={}
    df_dict['long']=long_flat
    df_dict['lat']=lat_flat
    df_dict['no2']=no2_flat
    df_dict['qa_value']=qa_value_flat
    df=pd.DataFrame(df_dict)
    
    file.close()
    return df
    

def merge_granules(sat_data_directory, file_name_list):
    df=pd.DataFrame(columns=['lat','long','no2','qa_value'])
    for f in file_name_list:
        data_file_path=os.path.join(sat_data_directory,f)
        temp_df=converter.sat_to_df(data_file_path)
        df=pd.concat([df,temp_df],ignore_index=True)
    
    return df


def df_to_gdf(df,gdf_crs):
    
    df['geometry']=df[["long","lat"]].values.tolist()
    #df['geometry']=df['geometry'].apply(Point)
    df['geometry']=geopandas.points_from_xy(x=df['long'], y=df['lat'])
    df['no2']=df['no2'].values.tolist()
    df['qa_value']=df['qa_value'].values.tolist()
    
    gdf=geopandas.GeoDataFrame(df,crs=gdf_crs,geometry='geometry')
    
    return gdf
 
def df_filtering(df ,qa_value_threshold):
    
    df=df[df.qa_value>qa_value_threshold]
    df=df.reset_index()
    df=df.drop_duplicates(subset=['long','lat'])
    
    return df

def clip_gdf(gdf,zone_corners_coordinates):

    cliped_gdf=geopandas.clip(gdf,zone_corners_coordinates)
    
    return cliped_gdf


def gdf_to_raster(gdf, raster_template, file_path):
    geom_value=((geom,value) for geom,value in zip(gdf.geometry, gdf['no2']))
    rasterized=rasterize(geom_value,\
                         out_shape=raster_template.shape,\
                             transform=raster_template.transform,
                             fill=-1E+300,\
                                 all_touched=True,\
                                     dtype=rasterio.float64,\
                                         merge_alg=rasterio.enums.MergeAlg('REPLACE'))

    with rasterio.open(file_path,\
                        'w', driver='GTiff',\
                            dtype=rasterio.float64,\
                                count=1,\
                                    width=raster_template.shape[1],\
                                        height=raster_template.shape[0],\
                                            transform=raster_template.transform) as dst:
         dst.write(rasterized,indexes=1)
         
    return rasterized
    


def file_list_generator(sat_data_directory, granule_end_date):
    file_name_list=[]
    for f in os.listdir(sat_data_directory):
        if f.endswith('.nc') and f[36:44]==granule_end_date:
            file_name_list.append(f)
    return file_name_list



def set_region(region):
    
    if region=='130w60w20n55n':
        return [-130,20,-60,55]
    else:
        raise "the region is not supported"

        
def convert_to_df(sat_data_directory, writing_directory, granule_end_date, write_to_file=False):
    
    file_name_list=converter.file_list_generator(sat_data_directory, granule_end_date)

    df=converter.merge_granules(sat_data_directory,file_name_list)
    
    if write_to_file:
        save_file_name=granule_end_date+'_df.pickle'
        save_path=os.path.join(writing_directory,save_file_name)
        df.to_pickle(save_path)
        
    return df


def convert_to_raster(sat_data_directory, writing_directory, raster_directory, resolution_list, granule_end_date, qa_value_threshold, region='130w60w20n55n'):
    #resolution options: '005' for 0.05 degree, '010' for 0.10 degree, '025' for 0.25 degree, '050' for 0.50 degree, '100' for 1.0 degree
    #region must be '130w60w20n55n'
    #granule end date format: yyyymmdd and string like '20230615'
    #writing directory: the total uncliped and unfilterd dataframe is saved (not geodataframe) and the final clipped and filtered tif (raster) files.
    #sat_data_directory: read sat data downloaded from NASA
    #qa_value_threshold: filter sat data based on qa_value (usually 0.5 or 0.75) below threshold is filtered out.
    #this saves raster to file.
    
    
    
    
    
    df=converter.convert_to_df(sat_data_directory,writing_directory, granule_end_date)
    
    
    df=converter.df_filtering(df,qa_value_threshold)
    gdf=converter.df_to_gdf(df,gdf_crs=4326)
    
    region_corener=converter.set_region(region)
    
    gdf=converter.clip_gdf(gdf,region_corener)
    gdf=gdf.reset_index()
    if not gdf.empty:
        for res in resolution_list:
            
            raster_name='raster_template_'+'130w60w20n55n'+'_'+res+'.tif'
            raster_file_name=os.path.join(raster_directory,raster_name)
            with rasterio.open(raster_file_name) as raster_template:
            #raster_file_name='lib/raster_templates/raster_template_'+'130w60w20n55n'+'_'+res+'.tif'
            #with pkg_resources.resource_stream(__name__, raster_file_name) as stream:
            #    raster_template=rasterio.open(stream)
            #raster_template=converter.get_raster_template('130w60w20n55n',res)
                new_raster_name=granule_end_date+'_'+res+'_'+str(qa_value_threshold)+'_raster.tif'
                new_raster=os.path.join(writing_directory, new_raster_name)
    
                rasterized=converter.gdf_to_raster(gdf,raster_template, new_raster)
        
                #raster_template.close()
            

    del df
    del gdf

def converter_run(read_directory,write_directory,raster_directory, granule_end_date, qa_value_threshold, df_print=False):
 
    #if df_print:
    #        converter.convert_to_df(read_directory,write_directory,granule_end_date,write_to_file=True)
        
    converter.convert_to_raster(read_directory,write_directory,raster_directory, ['005','010','025','050','100'], granule_end_date, qa_value_threshold)


